import os
from typing_extensions import Optional, Annotated, List

import httpx
import requests
import typer

from kleinkram.api_client import AuthenticatedClient
from kleinkram.error_handling import AccessDeniedException

file = typer.Typer(
    name="file",
    help="File operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@file.command("download")
def download_file(
    file_uuid: Annotated[List[str], typer.Option(help="UUIDs of the files")],
    local_path: Annotated[
        str,
        typer.Option(
            prompt=True,
            help="Local path to save the file",
        ),
    ],
):
    """
    Download files by UUIDs to a local path.\n
    Examples:\n
    klein file download --file-uuid="9d5a9..."  --file-uuid="9833f..." --local-path="~/Downloads" \n
    klein file download --file-uuid="9d5a9..."  --local-path="~/Downloads/example.bag"

    """
    client = AuthenticatedClient()
    url = f"/file/download"

    fixed_local_path = os.path.expanduser(local_path)

    isDir = os.path.isdir(fixed_local_path)
    chunk_size = 1024 * 100  # 100 KB chunks, adjust size if needed

    for file in file_uuid:
        response = client.get(
            url,
            params={"uuid": file, "expires": True},
        )
        if response.status_code >= 400:
            raise AccessDeniedException(
                f"Failed to download file: {response.json()['message']}",
                "Status Code: " + str(response.status_code),
            )
        download_url = response.text
        if isDir:
            filename = download_url.split("/")[6].split("?")[0]  # Trust me bro
            filepath = os.path.join(fixed_local_path, filename)
        elif not isDir and len(file_uuid) == 1:
            filepath = fixed_local_path
        else:
            raise ValueError("Multiple files can only be downloaded to a directory")
        if os.path.exists(filepath):
            raise FileExistsError(f"File already exists: {filepath}")
        print(f"Downloading to: {filepath}")
        filestream = requests.get(download_url, stream=True)
        with open(filepath, "wb") as f:
            for chunk in filestream.iter_content(chunk_size=chunk_size):
                if chunk:  # Filter out keep-alive new chunks
                    f.write(chunk)
            print(f"Completed")


@file.command("list")
def list_files(
    project: Optional[str] = typer.Option(None, help="Name of Project"),
    mission: Optional[str] = typer.Option(None, help="Name of Mission"),
    topics: Optional[str] = typer.Option(None, help="Comma separated list of topics"),
    tags: Optional[str] = typer.Option(
        None, help="Comma separated list of tagtype:tagvalue pairs"
    ),
):
    """
    List all files with optional filters for project, mission, or topics.

    Can list files of a project, mission, or with specific topics (Logical AND).
    Examples:\n
        - 'klein filelist'\n
        - 'klein file list --project "Project_1"'\n
        - 'klein file list --mission "Mission_1"'\n
        - 'klein file list --topics "/elevation_mapping/semantic_map,/elevation_mapping/elevation_map_raw"'\n
        - 'klein file list --topics "/elevation_mapping/semantic_map,/elevation_mapping/elevation_map_raw" --mission "Mission A"'
    """
    try:
        url = f"/file/filteredByNames"
        params = {}
        if project:
            params["projectName"] = project
        if mission:
            params["missionName"] = mission
        if topics:
            params["topics"] = topics
        if tags:
            params["tags"] = {}
            for tag in tags.split(","):
                tagtype, tagvalue = tag.split("§")
                params["tags"][tagtype] = tagvalue

        client = AuthenticatedClient()
        response = client.get(
            url,
            params=params,
        )
        if response.status_code >= 400:
            raise AccessDeniedException(
                f"Failed to fetch files: {response.json()['message']} ({response.status_code})",
                "Access Denied",
            )
        data = response.json()
        missions_by_project_uuid = {}
        files_by_mission_uuid = {}
        for file in data:
            mission_uuid = file["mission"]["uuid"]
            project_uuid = file["mission"]["project"]["uuid"]
            if project_uuid not in missions_by_project_uuid:
                missions_by_project_uuid[project_uuid] = []
            if mission_uuid not in missions_by_project_uuid[project_uuid]:
                missions_by_project_uuid[project_uuid].append(mission_uuid)
            if mission_uuid not in files_by_mission_uuid:
                files_by_mission_uuid[mission_uuid] = []
            files_by_mission_uuid[mission_uuid].append(file)

        print("Files by mission & Project:")
        for project_uuid, missions in missions_by_project_uuid.items():
            first_file = files_by_mission_uuid[missions[0]][0]
            print(f"* {first_file['mission']['project']['name']}")
            for mission in missions:
                print(f"  - {files_by_mission_uuid[mission][0]['mission']['name']}")
                for file in files_by_mission_uuid[mission]:
                    print(f"    - '{file['filename']}'")

    except httpx.HTTPError as e:
        print(f"Failed to fetch missions: {e}")
        raise e



import sys

import typer
from rich.console import Console
from rich.table import Table
from typing_extensions import Annotated

from kleinkram.api_client import AuthenticatedClient

user = typer.Typer(
    name="users",
    help="User operations",
    no_args_is_help=True,
)


@user.command("list")
def users():
    """List all users"""

    client = AuthenticatedClient()
    response = client.get("/user/all")
    response.raise_for_status()
    data = response.json()

    console = Console(file=sys.stderr)
    console_stdout = Console(file=sys.stdout)

    ####
    # Print user UUIDs for command piping
    ####
    console.print("\nUser UUIDs: ")
    for _user in data:
        console.print(" - ", end="")
        console_stdout.print(_user["uuid"])
    console.print("\n")

    ####
    # Print table of the users
    ####
    table = Table("UUID", "Name", "Email", "Role", expand=True, highlight=False)
    for _user in data:
        table.add_row(_user["uuid"], _user["name"], _user["email"], _user["role"])
    console.print(table)


@user.command("info")
def user_info():
    """Get logged in user info"""
    client = AuthenticatedClient()
    response = client.get("/user/me")
    response.raise_for_status()
    data = response.json()

    # print prettified user info
    console = Console()
    console.print(data)


@user.command("promote")
def promote(email: Annotated[str, typer.Option()]):
    """Promote another user to admin"""
    client = AuthenticatedClient()
    response = client.post("/user/promote", json={"email": email})
    response.raise_for_status()
    print("User promoted.")


@user.command("demote")
def demote(email: Annotated[str, typer.Option()]):
    """Demote another user from admin"""
    client = AuthenticatedClient()
    response = client.post("/user/demote", json={"email": email})
    response.raise_for_status()
    print("User demoted.")



import glob
import os
import queue
import re
import sys
import threading
from datetime import datetime
from functools import partial

import boto3
import tqdm
import typer
from boto3.s3.transfer import TransferConfig
from botocore.config import Config
from botocore.utils import calculate_md5
from rich import print
from rich.console import Console
from typing_extensions import Dict

from kleinkram.api_client import AuthenticatedClient


class TransferCallback:
    """
    Handle callbacks from the transfer manager.

    The transfer manager periodically calls the __call__ method throughout
    the upload process so that it can take action, such as displaying progress
    to the user and collecting data about the transfer.
    """

    def __init__(self):
        """
        Initialize the TransferCallback.

        This initializes an empty dictionary to hold progress bars for each file.
        """
        self._lock = threading.Lock()
        self.file_progress = {}

    def add_file(self, file_id, target_size):
        """
        Add a new file to track.

        :param file_id: A unique identifier for the file (e.g., file name or ID).
        :param target_size: The total size of the file being transferred.
        """
        with self._lock:
            tqdm_instance = tqdm.tqdm(
                total=target_size,
                unit="B",
                unit_scale=True,
                desc=f"Uploading {file_id}",
            )
            self.file_progress[file_id] = {
                "tqdm": tqdm_instance,
                "total_transferred": 0,
            }

    def __call__(self, file_id, bytes_transferred):
        """
        The callback method that is called by the transfer manager.

        Display progress during file transfer and collect per-thread transfer
        data. This method can be called by multiple threads, so shared instance
        data is protected by a thread lock.

        :param file_id: The identifier of the file being transferred.
        :param bytes_transferred: The number of bytes transferred in this call.
        """
        with self._lock:
            if file_id in self.file_progress:
                progress = self.file_progress[file_id]
                progress["total_transferred"] += bytes_transferred

                # Update tqdm progress bar
                progress["tqdm"].update(bytes_transferred)

    def close(self):
        """Close all tqdm progress bars."""
        with self._lock:
            for progress in self.file_progress.values():
                progress["tqdm"].close()


def create_transfer_callback(callback_instance, file_id):
    """
    Factory function to create a partial function for TransferCallback.
    :param callback_instance: Instance of TransferCallback.
    :param file_id: The unique identifier for the file.
    :return: A callable that can be passed as a callback to boto3's upload_file method.
    """
    return partial(callback_instance.__call__, file_id)


def expand_and_match(path_pattern):
    expanded_path = os.path.expanduser(path_pattern)
    expanded_path = os.path.expandvars(expanded_path)

    normalized_path = os.path.normpath(expanded_path)

    if "**" in normalized_path:
        file_list = glob.glob(normalized_path, recursive=True)
    else:
        file_list = glob.glob(normalized_path)

    return file_list


def uploadFiles(
    files_with_access: Dict[str, object], paths: Dict[str, str], nrThreads: int
):
    client = AuthenticatedClient()

    api_endpoint = client.tokenfile.endpoint
    if api_endpoint == "http://localhost:3000":
        minio_endpoint = "http://localhost:9000"
    else:
        minio_endpoint = api_endpoint.replace("api", "minio")

    _queue = queue.Queue()
    for file_with_access in files_with_access:
        _queue.put((file_with_access, str(paths[file_with_access["fileName"]])))

    threads = []
    transfer_callback = TransferCallback()

    for i in range(nrThreads):
        thread = threading.Thread(
            target=uploadFile,
            args=(_queue, minio_endpoint, transfer_callback),
        )
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()


def uploadFile(
    _queue: queue.Queue,
    minio_endpoint: str,
    transfer_callback: TransferCallback,
):
    config = Config(retries={"max_attempts": 10, "mode": "standard"})

    while True:
        try:
            file_with_access, filepath = _queue.get(timeout=3)

            if "error" in file_with_access and (
                file_with_access["error"] is not None or file_with_access["error"] != ""
            ):
                console = Console(file=sys.stderr, style="red", highlight=False)
                console.print(
                    f"Error uploading file: {file_with_access['fileName']} ({filepath}): {file_with_access['error']}"
                )
                _queue.task_done()
                continue

            access_key = file_with_access["accessCredentials"]["accessKey"]
            secret_key = file_with_access["accessCredentials"]["secretKey"]
            session_token = file_with_access["accessCredentials"]["sessionToken"]

            session = boto3.Session(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token,
            )

            s3 = session.resource("s3", endpoint_url=minio_endpoint, config=config)

            fileu_uid = file_with_access["fileUUID"]
            bucket = file_with_access["bucket"]

            transfer_config = TransferConfig(
                multipart_chunksize=10 * 1024 * 1024,
                max_concurrency=5,
            )
            with open(filepath, "rb") as f:
                md5_checksum = calculate_md5(f)
                file_size = os.path.getsize(filepath)
                transfer_callback.add_file(filepath, file_size)
                callback_function = create_transfer_callback(
                    transfer_callback, filepath
                )
                s3.Bucket(bucket).upload_file(
                    filepath,
                    fileu_uid,
                    Config=transfer_config,
                    Callback=callback_function,
                )

                client = AuthenticatedClient()
                res = client.post(
                    "/queue/confirmUpload",
                    json={"uuid": fileu_uid, "md5": md5_checksum},
                )
                res.raise_for_status()
            _queue.task_done()
        except queue.Empty:
            break
        except Exception as e:
            print("Error uploading file: " + filepath)
            print(e)
            _queue.task_done()


def canUploadMission(client: AuthenticatedClient, project_uuid: str):
    permissions = client.get("/user/permissions")
    permissions.raise_for_status()
    permissions_json = permissions.json()
    for_project = filter(
        lambda x: x["uuid"] == project_uuid, permissions_json["projects"]
    )
    max_for_project = max(map(lambda x: x["access"], for_project))
    return max_for_project >= 10


def promptForTags(setTags: Dict[str, str], requiredTags: Dict[str, str]):
    for required_tag in requiredTags:
        if required_tag["name"] not in setTags:
            while True:
                if required_tag["datatype"] in ["LOCATION", "STRING", "LINK"]:
                    tag_value = typer.prompt(
                        "Provide value for required tag " + required_tag["name"]
                    )
                    if tag_value != "":
                        break
                elif required_tag["datatype"] == "BOOLEAN":
                    tag_value = typer.confirm(
                        "Provide (y/N) for required tag " + required_tag["name"]
                    )
                    break
                elif required_tag["datatype"] == "NUMBER":
                    tag_value = typer.prompt(
                        "Provide number for required tag " + required_tag["name"]
                    )
                    try:
                        tag_value = float(tag_value)
                        break
                    except ValueError:
                        typer.echo("Invalid number format. Please provide a number.")
                elif required_tag["datatype"] == "DATE":
                    tag_value = typer.prompt(
                        "Provide date for required tag " + required_tag["name"]
                    )
                    try:
                        tag_value = datetime.strptime(tag_value, "%Y-%m-%d %H:%M:%S")
                        break
                    except ValueError:
                        print("Invalid date format. Please use 'YYYY-MM-DD HH:MM:SS'")

            setTags[required_tag["uuid"]] = tag_value


def is_valid_UUIDv4(uuid: str) -> bool:
    has_correct_length = len(uuid) == 36

    # is UUID4
    uuid_regex = (
        "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
    )
    is_valid_uuid = re.match(uuid_regex, uuid)

    return has_correct_length and is_valid_uuid


if __name__ == "__main__":
    res = expand_and_match(
        "~/Downloads/dodo_mission_2024_02_08-20240408T074313Z-003/**.bag"
    )
    print(res)



import importlib.metadata
import os
from datetime import datetime, timedelta
from enum import Enum

import httpx
import typer
from rich import print
from rich.table import Table
from typer.core import TyperGroup
from typer.models import Context
from typing_extensions import Annotated, List, Optional

from kleinkram.api_client import AuthenticatedClient
from kleinkram.auth.auth import login, setCliKey, logout
from kleinkram.endpoint.endpoint import endpoint
from kleinkram.error_handling import (
    ErrorHandledTyper,
    AccessDeniedException,
)
from kleinkram.file.file import file
from kleinkram.mission.mission import missionCommands
from kleinkram.project.project import project
from kleinkram.queue.queue import queue
from kleinkram.tag.tag import tag
from kleinkram.topic.topic import topic
from kleinkram.user.user import user
from .helper import (
    is_valid_UUIDv4,
    canUploadMission,
    promptForTags,
    expand_and_match,
    uploadFiles,
)


class CommandPanel(str, Enum):
    CoreCommands = "CORE COMMANDS"
    Commands = "COMMANDS"
    AdditionalCommands = "ADDITIONAL COMMANDS"


def version_callback(value: bool):
    if value:
        try:
            _version = importlib.metadata.version("kleinkram")
        except importlib.metadata.PackageNotFoundError:
            _version = "local"
        typer.echo(f"CLI Version: {_version}")
        raise typer.Exit()


class OrderCommands(TyperGroup):
    """

    The following code snippet is taken from https://github.com/tiangolo/typer/discussions/855 (see comment
    https://github.com/tiangolo/typer/discussions/855#discussioncomment-9824582) and adapted to our use case.
    """

    def list_commands(self, _ctx: Context) -> List[str]:
        order = list(CommandPanel)
        grouped_commands = {
            name: getattr(command, "rich_help_panel")
            for name, command in sorted(self.commands.items())
            if getattr(command, "rich_help_panel") in order
        }
        ungrouped_command_names = [
            command.name
            for command in self.commands.values()
            if command.name not in grouped_commands
        ]
        return [
            name
            for name, command in sorted(
                grouped_commands.items(),
                key=lambda item: order.index(item[1]),
            )
        ] + sorted(ungrouped_command_names)


app = ErrorHandledTyper(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    cls=OrderCommands,
    help=f"Kleinkram CLI\n\nThe Kleinkram CLI is a command line interface for Kleinkram. "
    f"For a list of available commands, run 'klein --help' or visit "
    f"https://docs.datasets.leggedrobotics.com/usage/cli/cli-getting-started.html for more information.",
)


@app.callback()
def version(
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    )
):
    pass


app.add_typer(project, rich_help_panel=CommandPanel.Commands)
app.add_typer(missionCommands, rich_help_panel=CommandPanel.Commands)

app.add_typer(topic, rich_help_panel=CommandPanel.Commands)
app.add_typer(file, rich_help_panel=CommandPanel.Commands)
app.add_typer(queue, rich_help_panel=CommandPanel.Commands)
app.add_typer(user, rich_help_panel=CommandPanel.Commands)
app.add_typer(tag, rich_help_panel=CommandPanel.Commands)
app.add_typer(endpoint, rich_help_panel=CommandPanel.AdditionalCommands)

app.command(rich_help_panel=CommandPanel.AdditionalCommands)(login)
app.command(rich_help_panel=CommandPanel.AdditionalCommands)(logout)
app.command(hidden=True)(setCliKey)


@app.command("download", rich_help_panel=CommandPanel.CoreCommands)
def download():
    print(
        "Not implemented yet. Consider using the 'klein file download' or 'klein mission download' commands."
    )


@app.command("upload", rich_help_panel=CommandPanel.CoreCommands, no_args_is_help=True)
def upload(
    path: Annotated[
        List[str],
        typer.Option(help="Path to files to upload, Regex supported"),
    ],
    project: Annotated[str, typer.Option(help="Name or UUID of a Project")],
    mission: Annotated[str, typer.Option(help="Name of UUID Mission to create")],
    tags: Annotated[
        Optional[List[str]],
        typer.Option(help="Tags to add to the mission"),
    ] = None,
    fix_filenames: Annotated[
        bool,
        typer.Option(help="Automatically fix filenames such that they are valid"),
    ] = False,
    create_project: Annotated[
        bool,
        typer.Option(help="Allows adding files to an existing mission"),
    ] = False,
    create_mission: Annotated[
        bool,
        typer.Option(help="Allows adding files to an existing mission"),
    ] = False,
    overwrite: Annotated[
        bool,
        typer.Option(
            help="Overwrite files with the same name.\n\n*WARNING:* This cannot be undone! This command will NOT delete"
            "converted files, i.g. if the file is of type 'some-name.bag' the converted 'some-name.mcap' file will not "
            "be deleted."
        ),
    ] = False,
    overwrite_all: Annotated[
        bool,
        typer.Option(
            help="Overwrite files with the same name.\n\n*WARNING:* This cannot be undone! This command WILL "
            "automatically delete converted files, i.g. if the file is of type 'some-name.bag' the converted "
            "'some-name.mcap' file will be deleted."
        ),
    ] = False,
):
    """
    Upload files matching the path to a mission in a project.

    The mission name must be unique within the project and not yet created.\n
    Multiple paths can be given by using the option multiple times.\n
    Examples:\n
        - 'klein upload --path "~/data/**/*.bag" --project "Project_1" --mission "Mission_1" --tags "0700946d-1d6a-4520-b263-0e177f49c35b:LEE-H" --tags "1565118d-593c-4517-8c2d-9658452d9319:Dodo"'\n

    """

    client = AuthenticatedClient()

    ##############################
    # Check if project exists
    ##############################
    if is_valid_UUIDv4(project):
        get_project_url = "/project/one"
        project_response = client.get(get_project_url, params={"uuid": project})
    else:
        get_project_url = "/project/byName"
        project_response = client.get(get_project_url, params={"name": project})

    if project_response.status_code >= 400:
        if not create_project and not is_valid_UUIDv4(project):
            raise AccessDeniedException(
                f"The project '{project}' does not exist or you do not have access to it.\n"
                f"Consider using the following command to create a project: 'klein project create' "
                f"or consider passing the flag '--create-project' to create the project automatically.",
                f"{project_response.json()['message']} ({project_response.status_code})",
            )
        elif is_valid_UUIDv4(project):
            raise ValueError(
                f"Project '{project}' does not exist. UUIDs cannot be used to create projects.\n"
                f"Please provide a valid project name or consider creating the project using the"
                f" following command: 'klein project create'"
            )
        else:
            print(f"Project '{project}' does not exist. Creating it now.")
            create_project_url = "/project/create"
            project_response = client.post(
                create_project_url,
                json={
                    "name": project,
                    "description": "autogenerated with kleinkram CLI",
                    "requiredTags": [],
                },
            )
            if project_response.status_code >= 400:
                msg = str(project_response.json()["message"])
                raise ValueError(
                    f"Failed to create project. Status Code: "
                    f"{str(project_response.status_code)}\n"
                    f"{msg}"
                )
            print("Project created successfully.")

    project_json = project_response.json()
    if not project_json["uuid"]:
        print(f"Project not found: '{project}'")
        return

    can_upload = canUploadMission(client, project_json["uuid"])
    if not can_upload:
        raise AccessDeniedException(
            f"You do not have the required permissions to upload to project '{project}'\n",
            "Access Denied",
        )

    if not tags:
        tags = []
    tags_dict = {item.split(":")[0]: item.split(":")[1] for item in tags}

    required_tags = (
        project_json["requiredTags"] if "requiredTags" in project_json else []
    )
    promptForTags(tags_dict, required_tags)

    ##############################
    # Check if mission exists
    ##############################
    if is_valid_UUIDv4(mission):
        get_mission_url = "/mission/one"
        mission_response = client.get(get_mission_url, params={"uuid": mission})
    else:
        get_mission_url = "/mission/byName"
        mission_response = client.get(get_mission_url, params={"name": mission})

    if mission_response.status_code >= 400:
        if not create_mission:
            raise AccessDeniedException(
                f"The mission '{mission}' does not exist or you do not have access to it.\n"
                f"Consider using the following command to create a mission: 'klein mission create' "
                f"or consider passing the flag '--create-mission' to create the mission automatically.",
                f"{mission_response.json()['message']} ({mission_response.status_code})",
            )
        else:
            print(f"Mission '{mission}' does not exist. Creating it now.")
            create_mission_url = "/mission/create"
            mission_response = client.post(
                create_mission_url,
                json={
                    "name": mission,
                    "projectUUID": project_json["uuid"],
                    "tags": tags_dict,
                },
            )
            if mission_response.status_code >= 400:
                raise ValueError(
                    f"Failed to create mission. Status Code: "
                    f"{str(mission_response.status_code)}\n"
                    f"{mission_response.json()['message'][0]}"
                )

    mission_json = mission_response.json()

    files = []
    for p in path:
        files.extend(expand_and_match(p))

    print(
        f"Uploading the following files to mission '{mission_json['name']}' in project '{project_json['name']}':"
    )
    filename_filepaths_map = {}
    for path in files:
        if not os.path.isdir(path):

            filename = path.split("/")[-1]
            filename_without_extension, extension = os.path.splitext(filename)
            if fix_filenames:

                # replace all non-alphanumeric characters with underscores
                filename_without_extension = "".join(
                    char if char.isalnum() else "_"
                    for char in filename_without_extension
                )

                # trim filename to 40 characters
                filename_without_extension = filename_without_extension[:40]
                filename = f"{filename_without_extension}{extension}"

            if (
                not filename.replace(".", "")
                .replace("_", "")
                .replace("-", "")
                .isalnum()
            ):
                raise ValueError(
                    f"Filename '{filename}' is not valid. It must only contain alphanumeric characters, underscores and "
                    f"hyphens. Consider using the '--fix-filenames' option to automatically fix the filenames."
                )

            if not 3 <= len(filename_without_extension) <= 40:
                raise ValueError(
                    f"Filename '{filename}' is not valid. It must be between 3 and 40 characters long. Consider using "
                    f"the '--fix-filenames' option to automatically fix the filenames."
                )

            filename_filepaths_map[filename] = path
            typer.secho(f" - {filename}", fg=typer.colors.RESET)
    print("\n\n")

    filenames = list(filename_filepaths_map.keys())

    if not filenames:
        raise ValueError("No files found matching the given path.")

    # validate filenames
    if len(filenames) != len(set(filenames)):
        raise ValueError(
            "Filenames must be unique. Please check the files you are trying to upload. This can happen if you have "
            "multiple files with the same name in different directories or use the '--fix-filenames' option."
        )

    # check if files already exist
    get_files_url = "/file/ofMission"
    response = client.get(
        get_files_url,
        params={"uuid": mission_json["uuid"]},
    )
    if response.status_code >= 400:
        raise ValueError(
            "Failed to check for existing files. Status Code: "
            + str(response.status_code)
            + "\n"
            + response.json()["message"]
        )

    existing_files = response.json()[0]
    conflicting_files = [
        file for file in existing_files if file["filename"] in filenames
    ]

    if conflicting_files and len(conflicting_files):
        print("The following files already exist in the mission:")
        for file in conflicting_files:
            typer.secho(f" - {file['filename']}", fg=typer.colors.RED, nl=False)
            if overwrite or overwrite_all:
                # delete existing files
                delete_files_url = f"/file/{file['uuid']}"
                response = client.delete(delete_files_url)
                if response.status_code >= 400:
                    raise ValueError(
                        "Failed to delete existing files. Status Code: "
                        + str(response.status_code)
                        + "\n"
                        + response.json()["message"]
                    )
                print("   » deleted")

                # check if converted files exist
                mcap_file = file["filename"].replace(".bag", ".mcap")

                if mcap_file == file["filename"]:
                    continue

                mcap_uuid = next(
                    (
                        file["uuid"]
                        for file in existing_files
                        if file["filename"] == mcap_file
                    ),
                    None,
                )

                if mcap_uuid and overwrite_all:
                    typer.secho(f"   {mcap_file}", fg=typer.colors.RED, nl=False)
                    delete_files_url = f"/file/{mcap_uuid}"
                    response = client.delete(delete_files_url)
                    if response.status_code >= 400:
                        raise ValueError(
                            "Failed to delete existing files. Status Code: "
                            + str(response.status_code)
                            + "\n"
                            + response.json()["message"]
                        )
                    print("  » deleted")
                elif mcap_uuid and not overwrite_all:
                    print(
                        f"   {mcap_file}  » skipped (consider using '--overwrite-all' to delete this file)"
                    )
                else:
                    print("   » not found")

            else:
                print("")

        if not overwrite and not overwrite_all:
            print(
                "\nYou may use the '--overwrite' or '--overwrite-all' flag to overwrite existing files."
            )
        print("")

    get_temporary_credentials = "/file/temporaryAccess"
    response = client.post(
        get_temporary_credentials,
        json={"filenames": filenames, "missionUUID": mission_json["uuid"]},
    )
    if response.status_code >= 400:
        raise ValueError(
            "Failed to upload data. Status Code: "
            + str(response.status_code)
            + "\n"
            + response.json()["message"][0]
        )

    uploadFiles(response.json(), filename_filepaths_map, 4)


@queue.command("list")
def list_queue():
    """List current Queue entities"""
    try:
        url = "/queue/active"
        startDate = datetime.now().date() - timedelta(days=1)
        client = AuthenticatedClient()
        response = client.get(url, params={"startDate": startDate})
        response.raise_for_status()
        data = response.json()
        table = Table("UUID", "filename", "mission", "state", "origin", "createdAt")
        for topic in data:
            table.add_row(
                topic["uuid"],
                topic["filename"],
                topic["mission"]["name"],
                topic["state"],
                topic["location"],
                topic["createdAt"],
            )
        print(table)

    except httpx.HTTPError as e:
        print(e)


@app.command("claim", hidden=True)
def claim():
    """
    Claim admin rights as the first user

    Only works if no other user has claimed admin rights before.
    """

    client = AuthenticatedClient()
    response = client.post("/user/claimAdmin")
    response.raise_for_status()
    print("Admin claimed.")


if __name__ == "__main__":
    app()



import httpx
import typer
from rich.console import Console
from rich.table import Table
from typing_extensions import Annotated

from kleinkram.api_client import AuthenticatedClient

project = typer.Typer(
    name="project",
    help="Project operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@project.command("list", help="List all projects")
def list_projects():
    """
    List all projects.
    """
    client = AuthenticatedClient()
    response = client.get("/project/filtered")
    response.raise_for_status()
    projects = response.json()[0]

    if len(projects) == 0:
        print("No projects found. Create a new project using 'klein project create'")
        return

    stdout_console = Console(stderr=False)
    stderr_console = Console(stderr=True)
    stderr_console.print(f"\nfound {len(projects)} projects with the following UUIDs:")

    # print the uuids to stdout for simple piping
    for p in projects:
        stderr_console.print(" - ", end="")
        stdout_console.print(p["uuid"])
    stderr_console.print("\n")

    # Print a summary table using rich to stderr
    table = Table(title="Projects", expand=True)
    table.add_column("Project UUID", width=10)
    table.add_column("Project Name", width=12)
    table.add_column("Description")
    for p in projects:
        table.add_row(p["uuid"], p["name"], p["description"])

    stderr_console.print(table)
    stderr_console.print("\n")


@project.command("details", help="Get details of a project", no_args_is_help=True)
def project_details(
    project_uuid: Annotated[
        str, typer.Argument(help="UUID of the project to get details of")
    ]
):
    """
    Get details of a project
    """
    client = AuthenticatedClient()
    response = client.get(f"/project/one?uuid={project_uuid}")
    response.raise_for_status()
    project = response.json()

    stdout_console = Console(stderr=False)
    stderr_console = Console(stderr=True)
    stderr_console.print(
        f"\nDetails of project with UUID {project_uuid}:", highlight=False
    )

    # Print the details to stderr using rich
    table = Table(title="Project Details", expand=True)
    table.add_column("Key", width=16)
    table.add_column("Value")
    for key, value in project.items():

        access_name_map = {0: "READ", 10: "CREATE", 20: "WRITE", 30: "DELETE"}

        if key == "project_accesses":
            value = ", ".join(
                [
                    f"'{access['accessGroup']['name']}' ({access_name_map[access['rights']]})"
                    for access in value
                ]
            )

        if key == "missions":
            value = ", ".join([f"'{mission['name']}'" for mission in value])

        if key == "creator":
            value = value["name"]

        table.add_row(key, f"{value}")

    stderr_console.print(table)
    stderr_console.print("\nList of missions:")
    for mission in project["missions"]:
        stderr_console.print(" - ", end="")
        stdout_console.print(mission["uuid"])


@project.command("create", no_args_is_help=True, help="Create a new project")
def create_project(
    name: Annotated[str, typer.Option(help="Name of Project")],
    description: Annotated[str, typer.Option(help="Description of Project")],
):
    """
    Create a new project
    """
    # Todo add required tags as option.
    try:
        url = "/project/create"
        client = AuthenticatedClient()
        response = client.post(
            url, json={"name": name, "description": description, "requiredTags": []}
        )  # TODO: Add required tags as option
        if response.status_code >= 400:
            response_json = response.json()
            response_text = response_json["message"]
            raise ValueError(f"Failed to create project\n » {response_text}!")

        stderr_console = Console(stderr=True)
        stderr_console.print(f"Project '{name}' created successfully.")

        stdout_console = Console(stderr=False)
        stderr_console.print("\nProject UUID:\n - ", end="")
        stdout_console.print(response.json()["uuid"])

    except httpx.HTTPError as e:
        print(f"Failed to create project: {e}")
        raise e


@project.command("delete", help="Delete a project")
def delete_project():
    raise NotImplementedError()



import os
import re

import httpx
import requests
import typer
from rich.console import Console
from rich.table import Table
from tqdm import tqdm
from typing_extensions import Annotated, Optional, List

from kleinkram.api_client import AuthenticatedClient
from kleinkram.error_handling import AccessDeniedException
from kleinkram.helper import expand_and_match, uploadFiles

missionCommands = typer.Typer(
    name="mission",
    help="Mission operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@missionCommands.command("tag")
def addTag(
    mission_uuid: Annotated[str, typer.Argument()],
    tagtype_uuid: Annotated[str, typer.Argument()],
    value: Annotated[str, typer.Argument()],
):
    """Tag a mission"""
    try:
        client = AuthenticatedClient()
        response = client.post(
            "/tag/addTag",
            json={"mission": mission_uuid, "tagType": tagtype_uuid, "value": value},
        )
        if response.status_code < 400:
            print("Tagged mission")
        else:
            print(response.json())
            print("Failed to tag mission")
            raise Exception("Failed to tag mission")
    except httpx.HTTPError as e:
        print(e)
        print("Failed to tag mission")
        raise e


@missionCommands.command("list")
def list_missions(
    project: Optional[str] = typer.Option(None, help="Name of Project"),
    table: Optional[bool] = typer.Option(
        True, help="Outputs a table with more information"
    ),
):
    """
    List all missions with optional filter for project.
    """

    url = "/mission"
    params = {}
    if project:
        url += f"/filteredByProjectName"
        params["projectName"] = project
    else:
        url += "/all"

    client = AuthenticatedClient()

    try:

        response = client.get(url, params=params)
        response.raise_for_status()

    except httpx.HTTPError:

        raise AccessDeniedException(
            f"Failed to fetch mission."
            f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
            f"{response.json()['message']} ({response.status_code})",
        )

    data = response.json()
    missions_by_project_uuid = {}
    for mission in data:
        project_uuid = mission["project"]["uuid"]
        if project_uuid not in missions_by_project_uuid:
            missions_by_project_uuid[project_uuid] = []
        missions_by_project_uuid[project_uuid].append(mission)

    if len(missions_by_project_uuid.items()) == 0:
        print(f"No missions found for project '{project}'. Does it exist?")
        return

    print("missions by Project:")
    if not table:
        for project_uuid, missions in missions_by_project_uuid.items():
            print(f"* {missions_by_project_uuid[project_uuid][0]['project']['name']}")
            for mission in missions:
                print(f"  - {mission['name']}")

    else:
        table = Table(
            "project",
            "name",
            "UUID",
            "creator",
            "createdAt",
            title="Missions",
            expand=True,
        )
        for project_uuid, missions in missions_by_project_uuid.items():
            for mission in missions:
                table.add_row(
                    mission["project"]["name"],
                    mission["name"],
                    mission["uuid"],
                    mission["creator"]["name"],
                    mission["createdAt"],
                )
        console = Console()
        console.print(table)


@missionCommands.command("byUUID")
def mission_by_uuid(
    uuid: Annotated[str, typer.Argument()],
    json: Optional[bool] = typer.Option(False, help="Output as JSON"),
):
    """
    Get mission name, project name, creator and table of its files given a Mission UUID

    Use the JSON flag to output the full JSON response instead.

    Can be run with API Key or with login.
    """
    url = "/mission/one"
    client = AuthenticatedClient()
    response = client.get(url, params={"uuid": uuid})

    try:
        response.raise_for_status()
    except httpx.HTTPError:
        raise AccessDeniedException(
            f"Failed to fetch mission. "
            f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
            f"{response.json()['message']} ({response.status_code})",
        )

    data = response.json()

    if json:
        print(data)
        return
    print(f"mission: {data['name']}")
    print(f"Creator: {data['creator']['name']}")
    print("Project: " + data["project"]["name"])
    table = Table("Filename", "Size", "date")

    if "files" not in data:
        print("No files found for mission.")
        return

    for file in data["files"]:
        table.add_row(file["filename"], f"{file['size']}", file["date"])
    console = Console()
    console.print(table)


@missionCommands.command("download")
def download(
    mission_uuid: Annotated[
        List[str], typer.Option(help="UUIDs of Mission to download")
    ],
    local_path: Annotated[str, typer.Option()],
    pattern: Optional[str] = typer.Option(
        None,
        help="Simple pattern to match the filename against. Allowed are alphanumeric characters,"
        " '_', '-', '.' and '*' as wildcard.",
    ),
):
    """

    Downloads all files of a mission to a local path.
    The local path must be an empty directory.

    """

    if not os.path.isdir(local_path):
        raise ValueError(f"Local path '{local_path}' is not a directory.")
    if not os.listdir(local_path) == []:

        full_local_path = os.path.abspath(local_path)

        raise ValueError(
            f"Local path '{full_local_path}' is not empty, it contains {len(os.listdir(local_path))} files. "
            f"The local target directory must be empty."
        )

    client = AuthenticatedClient()
    for single_mission_uuid in mission_uuid:
        response = client.get("/mission/download", params={"uuid": single_mission_uuid})
        try:
            response.raise_for_status()
        except httpx.HTTPError as e:
            raise AccessDeniedException(
                f"Failed to download file."
                f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
                f"{response.json()['message']} ({response.status_code})",
            )

        paths = response.json()
        if len(paths) == 0:
            continue

        # validate search pattern
        if pattern:
            if not re.match(r"^[a-zA-Z0-9_\-.*]+$", pattern):
                raise ValueError(
                    "Invalid pattern. Allowed are alphanumeric characters, '_', '-', '.' and '*' as wildcard."
                )

            regex = pattern.replace("*", ".*")
            pattern = re.compile(regex)

        print(f"Found {len(paths)} files in mission:")
        paths = [
            path for path in paths if not pattern or pattern.match(path["filename"])
        ]

        if pattern:
            print(
                f" » filtered to {len(paths)} files matching pattern '{pattern.pattern}'."
            )

        print(f"Start downloading {len(paths)} files to '{local_path}':\n")
        for path in paths:

            filename = path["filename"]

            response = requests.get(path["link"], stream=True)  # Enable streaming mode
            chunk_size = 1024 * 1024 * 10  # 10 MB chunks, adjust size if needed

            # Open the file for writing in binary mode
            with open(os.path.join(local_path, filename), "wb") as f:
                for chunk in tqdm(
                    response.iter_content(chunk_size=chunk_size),
                    unit="MB",
                    desc=filename,
                ):
                    if chunk:  # Filter out keep-alive new chunks
                        f.write(chunk)


@missionCommands.command("upload")
def upload(
    path: Annotated[
        List[str],
        typer.Option(prompt=True, help="Path to files to upload, Regex supported"),
    ],
    mission_uuid: Annotated[
        str, typer.Option(prompt=True, help="UUID of Mission to create")
    ],
):
    """
    Upload files matching the path to a mission in a project.

    The mission name must be unique within the project and already created.
    Multiple paths can be given by using the option multiple times.\n
    \n
    Examples:\n
        - 'klein upload --path "~/data/**/*.bag" --mission-uuid "2518cfc2-07f2-41a5-b74c-fdedb1b97f88" '\n

    """
    files = []
    for p in path:
        files.extend(expand_and_match(p))
    filenames = list(
        map(lambda x: x.split("/")[-1], filter(lambda x: not os.path.isdir(x), files))
    )
    if not filenames:
        raise ValueError("No files found matching the given path.")

    print(f"Uploading the following files to mission '{mission_uuid}':")
    filepaths = {}
    for path in files:
        if not os.path.isdir(path):
            filepaths[path.split("/")[-1]] = path
            typer.secho(f" - {path}", fg=typer.colors.RESET)

    try:
        client = AuthenticatedClient()
        get_temporary_credentials = "/file/temporaryAccess"
        response = client.post(
            get_temporary_credentials,
            json={"filenames": filenames, "missionUUID": mission_uuid},
        )
        if response.status_code >= 400:
            raise ValueError(
                "Failed to upload data. Status Code: "
                + str(response.status_code)
                + "\n"
                + response.json()["message"][0]
            )

        uploadFiles(response.json(), filepaths, 4)
    except Exception as e:
        print(e)
        print("Failed to upload files")
        raise e
